unit GWDropDemoAppInterface;

{---------------------------------------------------------------
Purpose: Implements App-to-TDropTarget interface for GWDropDemo app.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

TGWDropDemoAppInterface
-------------
Demos how to subclass TGWDropAppInterface and implement the
app-specific portion of an OLE-drop-accepting application.

Revision History
----------------
98-07-21 GW Renamed to GWDropDemoAppInterface for consistency
98-06-26 GW Original DropDemoAppInterface

To-do list
----------
-- Add format details block to Data_ContentToLines
-- Add save-to-file capability
 
---------------------------------------------------------------}

interface
uses
  Windows, ActiveX,
  Classes, forms,
  GWDropTarget;

Const
  FormatCountMax = 20;

Type
  //------------------------------------
  TGWDropRecord = record
  //------------------------------------
    FormatEtc: TFormatEtc;
    StgMedium: TStgMedium;
    datasize : integer;
  end;

  TGWDropRecordArray = array [0..FormatCountMax-1] of TGWDropRecord;

  //------------------------------------
  TGWDropDemoAppInterface = class(TGWDropAppInterface)
  //------------------------------------
  private
    //--------------------------------------------
    // App-specific data
    //--------------------------------------------
    // fields public due only to laziness...
    FDrop_StringSave: string;
    FFormatCount: integer;
    FDropRecord  : TGWDropRecordArray;

  public
    property FormatCount: integer read FFormatCount;
    property DropRecord:  TGWDropRecordArray read FDropRecord;

    constructor Create(AOwner: TForm; UseMonitor: Boolean); override;

    //---------------------------------------
    // Drop Event functions
    // We will override these since we want
    // fairly major change of function.
    //---------------------------------------

    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; override;

    function DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; override;

    function DragLeave: HResult; override;

    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; override;

    //----------------------------------
    // Drop Event Helpers
    //----------------------------------
    // Either not used or not overriden
    // function  Drop_Logic...
    // procedure Drop_Feedback...

    //--------------------------------------------
    // App-specific data and UI-handler functions
    //--------------------------------------------
    procedure KeyStateUpdate(grfKeyState: Longint);
    procedure EffectUpdate(dwEffect: Longint);
    procedure Data_Clear;
    procedure Data_HandleDrop(dataObj: IDataObject);
    procedure Data_ContentToLines(FmtNum: integer; L: TStrings; ByteLimit: integer);

    //--------------------------------------------
    //  Utility functions
    //--------------------------------------------
  end;

//--------------------------------
implementation
//--------------------------------
uses sysutils, stdctrls, graphics,
     GWDropDemo_Main;  // so we can fiddle with form directly

//--------------------------------------------
constructor TGWDropDemoAppInterface.Create(AOwner: TForm; UseMonitor: Boolean); // Monitor = true means use monitoring
//--------------------------------------------
Var N: integer;
Begin
  inherited Create(AOwner, UseMonitor);

  FFormatCount    := 0;
  For N := 0 to FormatCountMax - 1 do
  Begin
    FDropRecord[N].StgMedium.tymed := TYMED_NULL; // indicate not in use
    FDropRecord[N].datasize := 0;
    // don't care about FormatEtc
  end;
end;

//--------------------------------------------
function TGWDropDemoAppInterface.DragEnter(
            const dataObj: IDataObject;
              grfKeyState: Longint;
                       pt: TPoint;
              var dwEffect: Longint): HResult;
//--------------------------------------------
Var
  FormatEtc  : TFormatEtc;
  Rslt       : HResult;
  HDropAvail: Boolean;
  S: string;
Begin
  KeyStateUpdate(grfKeyState);
  fm_Main.lb_Effect.Caption   := '';
  // we're going to accept anything, but just for demo, we'll show
  // how to use QueryGetData
  With FormatEtc do
  Begin
    cfFormat := CF_HDROP;
    ptd      := nil;
    dwAspect := DVASPECT_CONTENT;
    lindex   := -1;
    tymed    := TYMED_HGLOBAL;
  end;

  Rslt := dataObj.QueryGetData(FormatEtc);
  HDropAvail := (Rslt = S_OK);

  // We could use this result to accept or reject the drop,
  // but instead we'll just use it for a bit of feedback

  If HDropAvail then
    S := 'Can drop, includes HDrop' else
    S := 'Can drop, no HDrop format';

  //-------------------------------------
  // Feedback
  //-------------------------------------
  FDrop_ColorSave  := (DropTarget.DropWinControl as TStaticText).Color;
  (DropTarget.DropWinControl as TStaticText).Color := clLime;

  FDrop_StringSave := (DropTarget.DropWinControl as TStaticText).Caption;
  (DropTarget.DropWinControl as TStaticText).Caption := S;

  FDrop_FeedbackOn := true;

  dwEffect := Drop_StdEffect(grfKeyState);
  EffectUpdate(dwEffect);
  Result := S_OK;
end;

//--------------------------------------------
function TGWDropDemoAppInterface.DragOver(
              grfKeyState: Longint;
                       pt: TPoint;
             var dwEffect: Longint): HResult;
//--------------------------------------------
Begin
  KeyStateUpdate(grfKeyState);
  fm_Main.lb_Effect.Caption   := '';

  // No particular data processing
  dwEffect := Drop_StdEffect(grfKeyState);

  // No particular feedback

  EffectUpdate(dwEffect);
  Result := S_OK;
end;

//--------------------------------------------
function TGWDropDemoAppInterface.DragLeave: HResult;
//--------------------------------------------
Begin

  //-------------------------------------
  // Feedback
  //-------------------------------------
  if FDrop_FeedbackOn then
  Begin
    (DropTarget.DropWinControl as TStaticText).Color   := FDrop_ColorSave;
    (DropTarget.DropWinControl as TStaticText).Caption := FDrop_StringSave;
  end;
  FDrop_FeedbackOn := false;

  Result := S_OK;
end;

//--------------------------------------------
function TGWDropDemoAppInterface.Drop(
            const dataObj: IDataObject;
              grfKeyState: Longint;
                       pt: TPoint;
             var dwEffect: Longint): HResult;
//--------------------------------------------
Begin
  KeyStateUpdate(grfKeyState);
  fm_Main.lb_Effect.Caption   := '';
  //-------------------------------------
  // Deal with the data
  //-------------------------------------
  Data_HandleDrop(dataObj);

  //-------------------------------------
  // Feedback
  //-------------------------------------
  if FDrop_FeedbackOn then
  Begin
    (DropTarget.DropWinControl as TStaticText).Color   := FDrop_ColorSave;
    (DropTarget.DropWinControl as TStaticText).Caption := FDrop_StringSave;
  end;
  FDrop_FeedbackOn := false;

  if (FOwnerForm as Tfm_Main).SafeMode then
  Begin
    dwEffect := DROPEFFECT_COPY; // so we tell source app not to delete
  end else
  Begin
    dwEffect := Drop_StdEffect(grfKeyState);
  end;

  EffectUpdate(dwEffect);
  Result := S_OK;
end;

//--------------------------------------------
procedure KeyInfoUpdate;
//--------------------------------------------
Begin

end;
//--------------------------------------------
procedure TGWDropDemoAppInterface.Data_Clear;
//--------------------------------------------
Var N: integer;
Begin
  FFormatCount    := 0;
  For N := 0 to FormatCountMax - 1 do
  Begin
    If (FDropRecord[N].StgMedium.tymed <> TYMED_NULL) and
       (FDropRecord[N].StgMedium.unkForRelease = nil) then { we're responsible for freeing }
       ReleaseStgMedium(FDropRecord[N].StgMedium);

    FDropRecord[N].StgMedium.tymed := TYMED_NULL; // indicate not in use
  end;
end;

//-------------------------------------------------------
procedure TGWDropDemoAppInterface.Data_HandleDrop(dataObj: IDataObject);
//-------------------------------------------------------
Var
  Rslt: HResult;
  EnumFormatEtc: IEnumFormatEtc;
  FetchedCount: longint;
  FE: TFormatEtc;
  N: integer;
  Handle: HGlobal;
Const FetchRequest = 1;
Begin
  Data_Clear;

  //---------------------------------------
  // Get all the EnumFormats
  //---------------------------------------
  Rslt := dataObj.EnumFormatEtc(DATADIR_GET, EnumFormatEtc);
  FFormatCount := 0;
  Rslt := EnumFormatEtc.Reset;
  Rslt := S_OK;
  While (Rslt = S_OK) and (FFormatCount < FormatCountMax-1) do
  Begin
    Rslt := EnumFormatEtc.Next(FetchRequest, FE, @FetchedCount);
    If Rslt = S_OK then
    Begin
      FDropRecord[FFormatCount].FormatEtc.cfFormat := FE.cfFormat;
      FDropRecord[FFormatCount].FormatEtc.ptd      := FE.ptd;
      FDropRecord[FFormatCount].FormatEtc.dwAspect := FE.dwAspect;
      FDropRecord[FFormatCount].FormatEtc.lindex   := FE.lindex;
      FDropRecord[FFormatCount].FormatEtc.tymed    := FE.tymed;
      Inc(FFormatCount);
    end;
  end;

  //-------------------------------------------------------------
  // Get all HGLOBAL data (StgMediums with global-allocated data)
  //-------------------------------------------------------------
  For N := 0 to FFormatCount - 1 do
  Begin
    If (DropRecord[N].FormatEtc.tymed and TYMED_HGLOBAL) <> 0 then
    Begin
      FE.cfFormat  :=  FDropRecord[N].FormatEtc.cfFormat;
      FE.ptd       :=  FDropRecord[N].FormatEtc.ptd     ;
      FE.dwAspect  :=  FDropRecord[N].FormatEtc.dwAspect;
      FE.lindex    :=  FDropRecord[N].FormatEtc.lindex;
      FE.tymed     :=  TYMED_HGLOBAL;
      Rslt := dataObj.GetData(FE, FDropRecord[N].StgMedium);
      If Rslt = S_OK then
      Begin
        Handle := FDropRecord[N].StgMedium.hGlobal;
        FDropRecord[N].datasize := GlobalSize(Handle);
      end;
    end;
  end;

  //-------------------------------------------------------------
  // Tell form to update list and memo
  //-------------------------------------------------------------
  (FOwnerForm as Tfm_Main).OnOLEDrop;
end;

//-------------------------------
procedure TGWDropDemoAppInterface.Data_ContentToLines(
          FmtNum: integer; L: TStrings; ByteLimit: integer);
//-------------------------------
Var
  FmtName: String;
  S, S1, S2, S3: string;
  C: char;
  CharNum, datasize, CharCountThisLine: integer;
  ptr: PChar;
Const
  CharPerLine = 16;
  MaxDataSize = 4096;

  procedure AddLine;
  Begin
    While Length(S2) < 48 do S2 := S2 + ' ';
    S := S1 + ' | ' + S2 + ' | ' + S3;
    L.Add(S);
    S1 := IntToHex(CharNum+1,6);  S2 := '';  S3 := '';
    CharCountThisLine := 0;
  end;

Label the_end;
Begin
  If (FmtNum < 0) or (FmtNum > FormatCountMax-1) then
  Begin
    S := 'Error: Attempt to add format: ' + IntToStr(FmtNum);
    L.Add(S);
    goto the_end;
  end;

  FmtName := cfFormatCaption(DropRecord[FmtNum].FormatEtc.cfFormat);
  L.Add('Format: '+ FmtName);

  If DropRecord[FmtNum].StgMedium.tymed <> TYMED_HGLOBAL then
  Begin
    S := 'Format does not use a global memory handle.';
    L.Add(S);
    goto the_end;
  end;

  datasize := DropRecord[FmtNum].datasize;
  ptr      := GlobalLock(DropRecord[FmtNum].StgMedium.HGlobal);
  try
    CharNum           := 0;
    CharCountThisLine := 0;
    S1 := IntToHex(CharNum,6);  S2 := '';  S3 := '';
    While (CharNum < datasize) and (CharNum < ByteLimit) do
    Begin
      Inc(CharCountThisLine);
      C := ptr[CharNum];
      S2 := S2 + ' ' + IntToHex(Ord(C),2);
      If (C >= ' ') and (Ord(C) <= $7F) then S3 := S3 + C else S3 := S3 + '.';

      If CharCountThisLine = CharPerLine then AddLine;
      Inc(CharNum);
    end;
    If CharCountThisLine > 0 then AddLine;
    If CharNum = ByteLimit then L.Add('...more...');
  finally
    GlobalUnLock(DropRecord[FmtNum].StgMedium.HGlobal);
  end;
the_end:
end;

//--------------------------------------------------
procedure TGWDropDemoAppInterface.KeyStateUpdate(grfKeyState: Longint);
//--------------------------------------------------
Var S: string;
  procedure Add(SAdd: string);
  Begin
    If S <> '' then S := S + '+';
    S := S + SAdd;
  end;
Const
  MK_ALT = $020;  // this is not defined with the others for some reason

Begin
  S := '';
  If (grfKeyState and MK_SHIFT  ) <> 0 then Add('Sh');
  If (grfKeyState and MK_CONTROL) <> 0 then Add('Ctl');
  If (grfKeyState and MK_ALT    ) <> 0 then Add('Alt');
  If (grfKeyState and MK_LBUTTON) <> 0 then Add('ML');
  If (grfKeyState and MK_MBUTTON) <> 0 then Add('MM');
  If (grfKeyState and MK_RBUTTON) <> 0 then Add('MR');
  fm_Main.lb_KeyState.Caption := IntToHex(grfKeyState, 8)+': '+S;
end;

//--------------------------------------------------
procedure TGWDropDemoAppInterface.EffectUpdate(dwEffect: Longint);
//--------------------------------------------------
Var S: string;
  procedure Add(SAdd: string);
  Begin
    If S <> '' then S := S + '+';
    S := S + SAdd;
  end;
Const
  MK_ALT = $020;  // this is not defined with the others for some reason

Begin
  S := '';
  // If (grfKeyState and DROPEFFECT_NONE  ) <> 0 then Add('--');
  If (dwEffect and DROPEFFECT_COPY  ) <> 0 then Add('Copy');
  If (dwEffect and DROPEFFECT_MOVE  ) <> 0 then Add('Move');
  If (dwEffect and DROPEFFECT_LINK  ) <> 0 then Add('Link');
  If (dwEffect and DROPEFFECT_SCROLL) <> 0 then Add('Scr');
  fm_Main.lb_Effect.Caption := IntToHex(dwEffect, 8)+': '+S;
end;

end.

